@@ -100,18 +100,25 @@
   (declare (type lvar lvar))
   (ir2-lvar-primitive-type (lvar-info lvar)))
 
+;;; Return true if a constant LEAF is of a type which we can legally
+;;; directly reference in code. Named constants with arbitrary pointer
+;;; values cannot, since we must preserve EQLness.
+(defun legal-immediate-constant-p (leaf)
+  (declare (type constant leaf))
+  (or (not (leaf-has-source-name-p leaf))
+      (sb-xc:typep (constant-value leaf) '(or symbol number character))))
+
 ;;; If LVAR is used only by a REF to a leaf that can be delayed, then
 ;;; return the leaf, otherwise return NIL.
 (defun lvar-delayed-leaf (lvar)
   (declare (type lvar lvar))
-  (unless (lvar-dynamic-extent lvar)
-    (let ((use (lvar-uses lvar)))
-      (and (ref-p use)
-           (let ((leaf (ref-leaf use)))
-             (etypecase leaf
-               (lambda-var (if (null (lambda-var-sets leaf)) leaf nil))
-               (constant leaf)
-               ((or functional global-var) nil)))))))
+  (let ((use (lvar-uses lvar)))
+    (and (ref-p use)
+         (let ((leaf (ref-leaf use)))
+           (etypecase leaf
+             (lambda-var (if (null (lambda-var-sets leaf)) leaf nil))
+             (constant leaf)
+             ((or functional global-var) nil))))))
 
 ;;; Annotate a normal single-value lvar. If its only use is a ref that
 ;;; we are allowed to delay the evaluation of, then we mark the lvar
@@ -126,10 +133,7 @@
       (setf (ir2-lvar-kind info) :delayed))
      (t (let ((tn (make-normal-tn (ir2-lvar-primitive-type info))))
           (setf (tn-type tn) (lvar-type lvar))
-          (setf (ir2-lvar-locs info) (list tn))
-          (when (lvar-dynamic-extent lvar)
-            (setf (ir2-lvar-stack-pointer info)
-                  (make-stack-pointer-tn)))))))
+          (setf (ir2-lvar-locs info) (list tn))))))
   (ltn-annotate-casts lvar)
   (values))
 
@@ -148,7 +152,6 @@
 ;;; reference, otherwise we annotate for a single value.
 (defun annotate-fun-lvar (lvar &optional (delay t))
   (declare (type lvar lvar))
-  (aver (not (lvar-dynamic-extent lvar)))
   (let* ((tn-ptype (primitive-type (lvar-type lvar)))
          (info (make-ir2-lvar tn-ptype)))
     (setf (lvar-info lvar) info)
@@ -162,24 +165,43 @@
   (values))
 
 #+call-symbol
-(defoptimizer (%coerce-callable-for-call ltn-annotate) ((fun) node ltn-policy)
-  (declare (ignore ltn-policy))
-  (let ((dest (node-dest node)))
+(defoptimizer (%coerce-callable-for-call ltn-annotate) ((fun) node)
+  (multiple-value-bind (dest dest-lvar)
+      (and (node-lvar node)
+           (principal-lvar-end (node-lvar node)))
     (cond ((and (basic-combination-p dest)
                 (eq (basic-combination-kind dest) :full)
-                (eq (lvar-uses (basic-combination-fun dest)) node)
+                (eq (basic-combination-fun dest) dest-lvar)
                 ;; Everything else can't handle NIL, just don't
                 ;; bother optimizing it.
-                (not (and (constant-lvar-p fun)
-                          (null (lvar-value fun)))))
+                (not (lvar-value-is fun nil)))
            (setf (basic-combination-fun dest) fun
-                 (basic-combination-args node) '(nil)
+                 (basic-combination-args node) '()
+                 (lvar-dest fun) dest
                  (node-lvar node) nil
                  (lvar-info fun) (make-ir2-lvar (primitive-type (lvar-type fun))))
            (annotate-1-value-lvar fun))
           (t
            (ltn-default-call node)))))
 
+(defun unboxed-specialized-return-p (name)
+  (and (typep name '(cons (eql sb-impl::specialized-xep)))
+       (let* ((specifier (cadddr name))
+              (type (and (neq specifier '*)
+                         (values-specifier-type (cadddr name)))))
+         (and (values-type-p type)
+              (not (or (values-type-optional type)
+                       (values-type-rest type)))
+              type))))
+
+(defun unboxed-return-p (combination)
+  (let ((fun (basic-combination-fun combination)))
+    (or (let ((info (basic-combination-fun-info combination)))
+          (and info
+               (ir1-attributep (fun-info-attributes info) unboxed-return)
+               (fun-type-returns (lvar-type fun))))
+        (unboxed-specialized-return-p (lvar-fun-name fun)))))
+
 ;;; If TAIL-P is true, then we check to see whether the call can
 ;;; really be a tail call by seeing if this function's return
 ;;; convention is :UNKNOWN. If so, we move the call block successor
@@ -189,9 +211,13 @@
 (defun flush-full-call-tail-transfer (call)
   (declare (type basic-combination call))
   (let ((tails (and (node-tail-p call)
-                    (lambda-tail-set (node-home-lambda call)))))
+                    (lambda-tail-set (node-home-lambda call))))
+        (unboxed-return (unboxed-return-p call)))
     (cond ((not tails))
-          ((eq (return-info-kind (tail-set-info tails)) :unknown)
+          ((eq (return-info-kind (tail-set-info tails))
+               (if unboxed-return
+                   :unboxed
+                   :unknown))
            (ir2-change-node-successor call
                                       (component-tail (block-component (node-block call)))))
           (t
@@ -199,9 +225,10 @@
   (values))
 
 (defun signal-delayed-combination-condition (call)
-  (let ((*compiler-error-context* call)
-        (delayed (combination-info call)))
-    (apply #'funcall delayed)))
+  (let* ((delayed (combination-info call))
+         (*compiler-error-context* call))
+    (when (consp delayed)
+      (apply #'funcall delayed))))
 
 ;;; We set the kind to :FULL or :FUNNY, depending on whether there is
 ;;; an IR2-CONVERT method. If a funny function, then we inhibit tail
@@ -216,6 +243,8 @@
   (let ((kind (basic-combination-kind call))
         (info (basic-combination-fun-info call)))
 
+    (rewrite-full-call call)
+
     (dolist (arg (basic-combination-args call))
       (unless (lvar-info arg)
         (setf (lvar-info arg)
@@ -233,8 +262,7 @@
          (when (basic-combination-info call)
            (signal-delayed-combination-condition call))
          (setf (basic-combination-kind call) :full))
-       (setf (basic-combination-info call) :full)
-       (rewrite-full-call call))))
+       (setf (basic-combination-info call) :full))))
   (annotate-fun-lvar (basic-combination-fun call))
   (values))
 
@@ -285,11 +313,7 @@
       (loop for type in lvar-types
             for tn in (ir2-lvar-locs info)
             do (setf (tn-type tn) type)))
-    (setf (lvar-info lvar) info)
-    (when (lvar-dynamic-extent lvar)
-      (aver (proper-list-of-length-p types 1))
-      (setf (ir2-lvar-stack-pointer info)
-            (make-stack-pointer-tn))))
+    (setf (lvar-info lvar) info))
   (ltn-annotate-casts lvar)
   (values))
 
@@ -376,13 +400,13 @@
                       (primitive-types))
               (let ((n-values (nth-value 1 (values-types
                                             (lvar-derived-type arg)))))
-                (loop for (prim-type . lvar-type) = (pop types)
-                      repeat n-values
+                (loop repeat n-values
                       do
-                      (primitive-types (or prim-type
-                                           *backend-t-primitive-type*))
-                      (lvar-types (or lvar-type
-                                      *universal-type*)))
+                      (destructuring-bind (&optional prim-type . lvar-type) (pop types)
+                        (primitive-types (or prim-type
+                                             *backend-t-primitive-type*))
+                        (lvar-types (or lvar-type
+                                        *universal-type*))))
                 (annotate-fixed-values-lvar
                  arg
                  (primitive-types) (lvar-types))))))))
@@ -451,6 +475,10 @@
   (annotate-ordinary-lvar (set-value node))
   (values))
 
+(defoptimizer (%%primitive ltn-annotate) ((template &rest args) node)
+  (ltn-default-call node)
+  (setf (basic-combination-info node) (lvar-value template)))
+
 ;;; If the only use of the TEST lvar is a combination annotated with a
 ;;; conditional template, then don't annotate the lvar so that IR2
 ;;; conversion knows not to emit any code, otherwise annotate as an
@@ -463,12 +491,19 @@
   (let* ((test (if-test node))
          (use (lvar-uses test)))
     (unless (and (combination-p use)
+                 (immediately-used-p test use)
                  (let ((info (basic-combination-info use)))
                    (and (template-p info)
                         (template-conditional-p info))))
       (annotate-ordinary-lvar test)))
   (values))
 
+(defun ltn-analyze-jump-table (node)
+  (declare (type jump-table node))
+  (setf (node-tail-p node) nil)
+  (annotate-ordinary-lvar (jump-table-index node))
+  (values))
+
 ;;; If there is a value lvar, then annotate it for unknown values. In
 ;;; this case, the exit is non-local, since all other exits are
 ;;; deleted or degenerate by this point.
@@ -476,9 +511,28 @@
   (setf (node-tail-p node) nil)
   (let ((value (exit-value node)))
     (when value
-      (annotate-unknown-values-lvar value)))
+      (if (lvar-single-value-p (node-lvar node))
+          (annotate-fixed-values-lvar value (list *backend-t-primitive-type*))
+          (annotate-unknown-values-lvar value))))
   (values))
 
+;;; Annotate DYNAMIC-EXTENT's info lvar as a stack lvar, which cleanup
+;;; code will use to clean up any values stack allocated in this
+;;; extent. Stack analysis will insert code to initialize this lvar
+;;; wherever necessary.
+(defun ltn-analyze-dynamic-extent (node)
+  (declare (type cdynamic-extent node))
+  (let ((lvar (dynamic-extent-info node))
+        (2comp (component-info (node-component node))))
+    (when (and lvar (not (lvar-info lvar)))
+      (setf (ir2-component-stack-allocates-p 2comp) t)
+      (setf (lvar-dest lvar) node)
+      (let ((info (make-ir2-lvar *backend-t-primitive-type*)))
+        (setf (lvar-info lvar) info)
+        (setf (ir2-lvar-kind info) :stack)
+        (setf (ir2-lvar-locs info)
+              (list (make-stack-pointer-tn)))))))
+
 ;;; We need a special method for %UNWIND-PROTECT that ignores the
 ;;; cleanup function. We don't annotate either arg, since we don't
 ;;; need them at run-time.
@@ -486,24 +540,16 @@
 ;;; (The default is o.k. for %CATCH, since environment analysis
 ;;; converted the reference to the escape function into a constant
 ;;; reference to the NLX-INFO.)
-(defoptimizer (%unwind-protect ltn-annotate) ((escape cleanup)
-                                              node
-                                              ltn-policy)
-  (declare (ignore escape cleanup ltn-policy))
+(defoptimizer (%unwind-protect ltn-annotate) ((escape cleanup) node)
   (setf (basic-combination-info node) :funny)
   (setf (node-tail-p node) nil))
 
 ;;; Make sure that arguments of magic functions are not annotated.
 ;;; (Otherwise the compiler may dump its internal structures as
 ;;; constants :-()
-(defoptimizer (%pop-values ltn-annotate) ((%lvar) node ltn-policy)
-  (declare (ignore %lvar node ltn-policy)))
-(defoptimizer (%dummy-dx-alloc ltn-annotate) ((target source) node ltn-policy)
-  (declare (ignore target source node ltn-policy)))
+(defoptimizer (%pop-values ltn-annotate) ((%lvar)))
 
-(defoptimizer (%nip-values ltn-annotate) ((&rest lvars)
-                                          node ltn-policy)
-  (declare (ignore node ltn-policy))
+(defoptimizer (%nip-values ltn-annotate) ((&rest lvars))
   ;; Undo the optimization performed by LTN-ANALYZE-MV-CALL,
   ;; which only uses the CSP of the first argument.
   (loop for lvar-lvar in lvars
@@ -542,7 +588,13 @@
                       (funcall (second type) value)
                       (sb-xc:typep value type))))
           (cond (lvar
-                 (and (constant-lvar-p lvar)
+                 (and (if (policy (let ((uses (lvar-uses lvar)))
+                                    (if (consp uses)
+                                        (car uses)
+                                        uses))
+                              (= preserve-constants 3))
+                          (constant-lvar-ignore-types-p lvar nil)
+                          (constant-lvar-p lvar))
                       (type-p (lvar-value lvar) (cdr restr))))
                 (tn
                  (and (eq (tn-kind tn) :constant)
@@ -574,6 +626,26 @@
         (unless (operand-restriction-ok type (lvar-ptype arg)
                                         :lvar arg)
           (return nil))))))
+(defun diagnose-template-args (template call)
+  ;; Scan all except the MORE args. If you've managed to create not-ok MORE args,
+  ;; you're probably smart enough to figure it out on your own.
+  (let ((args (basic-combination-args call))
+        (types (template-arg-types template)))
+    (unless (= (length args) (length types))
+      (return-from diagnose-template-args "bad length"))
+    (do ((args args (cdr args))
+         (i 0 (1+ i))
+         (any-fail nil)
+         (types types (cdr types)))
+        ((null types) any-fail)
+      (let* ((arg (car args))
+             (type (car types))
+             (ok (operand-restriction-ok type (lvar-ptype arg) :lvar arg)))
+        (let ((*print-pretty* nil))
+          (format t "arg~d: is ~s need ~s, ~a~%" i
+                  (primitive-type-name (lvar-ptype arg))
+                  type (if ok "OK" "FAIL")))
+        (unless ok (setq any-fail :fail))))))
 
 ;;; Check that TEMPLATE can be used with the specifed RESULT-TYPE.
 ;;; Result type checking is pretty different from argument type
@@ -583,7 +655,7 @@
 ;;; before we run out of restrictions, then we only succeed if the
 ;;; leftover restrictions are *. If we run out of restrictions before
 ;;; we run out of result types, then we always win.
-(defun template-results-ok (template result-type)
+(defun template-results-ok (template result-type &optional single-value-p)
   (declare (type template template)
            (type ctype result-type))
   (when (template-more-results-type template)
@@ -591,19 +663,28 @@
   (let ((types (template-result-types template)))
     (cond
      ((values-type-p result-type)
-      (do ((ltypes (append (args-type-required result-type)
-                           (args-type-optional result-type))
-                   (rest ltypes))
-           (types types (rest types)))
-          ((null ltypes)
-           (dolist (type types t)
-             (unless (eq type '*)
-               (return nil))))
-        (when (null types) (return t))
-        (let ((type (first types)))
-          (unless (operand-restriction-ok type
-                                          (primitive-type (first ltypes)))
-            (return nil)))))
+      (if (and single-value-p
+               (let ((optional (vop-info-optional-results template)))
+                 (and optional
+                      (not (eql (car optional) 0))
+                      (= (length optional)
+                         (1- (length types))))))
+          (operand-restriction-ok (car types)
+                                  (primitive-type (or (car (args-type-required result-type))
+                                                      (car (args-type-optional result-type)))))
+          (do ((ltypes (append (args-type-required result-type)
+                               (args-type-optional result-type))
+                       (rest ltypes))
+               (types types (rest types)))
+              ((null ltypes)
+               (dolist (type types t)
+                 (unless (eq type '*)
+                   (return nil))))
+            (when (null types) (return t))
+            (let ((type (first types)))
+              (unless (operand-restriction-ok type
+                                              (primitive-type (first ltypes)))
+                (return nil))))))
      (types
       (operand-restriction-ok (first types) (primitive-type result-type)))
      (t t))))
@@ -629,7 +710,7 @@
   (let* ((guard (template-guard template))
          (lvar (node-lvar call))
          (dtype (node-derived-type call)))
-    (cond ((and guard (not (funcall guard)))
+    (cond ((and guard (not (funcall guard call)))
            (values nil :guard))
           ((not (template-args-ok template call safe-p))
            (values nil
@@ -637,12 +718,13 @@
                        :arg-check
                        :arg-types)))
           ((template-conditional-p template)
-           (let ((dest (lvar-dest lvar)))
-             (if (and (if-p dest)
-                      (immediately-used-p (if-test dest) call))
-                 (values t nil)
-                 (values nil :conditional))))
-          ((template-results-ok template dtype)
+           (or (vop-existsp :named sb-vm::move-conditional-result)
+               (let ((dest (lvar-dest lvar)))
+                 (if (and (if-p dest)
+                          (immediately-used-p (if-test dest) call))
+                     (values t nil)
+                     (values nil :conditional)))))
+          ((template-results-ok template dtype (lvar-single-value-p (node-lvar call)))
            (values t nil))
           (t
            (values nil :result-types)))))
@@ -794,9 +876,9 @@
                             (template-or-lose 'call-named)))
                        *efficiency-note-cost-threshold*)))
       (dolist (try (fun-info-templates (basic-combination-fun-info call)))
-        (when (> (template-cost try) max-cost) (return)) ; FIXME: UNLESS'd be cleaner.
+        (when (> (template-cost try) max-cost) (return))
         (let ((guard (template-guard try)))
-          (when (and (or (not guard) (funcall guard))
+          (when (and (or (not guard) (funcall guard call))
                      (or (not safe-p)
                          (ltn-policy-safe-p (template-ltn-policy try)))
                      (not (and (eq ltn-policy :safe)
@@ -884,11 +966,11 @@
       ;; transforms or VOPs or whatever.
       (unless template
         (ltn-default-call call)
-        (when (let ((funleaf (physenv-lambda (node-physenv call)))
+        (when (let ((funleaf (environment-lambda (node-environment call)))
                     (name (lvar-fun-name (combination-fun call))))
                 (and (leaf-has-source-name-p funleaf)
                      (eq name (leaf-source-name funleaf))
-                     (not (sb-vm::static-fdefn-offset name))
+                     (not (static-fdefn-p name))
                      (let ((info (basic-combination-fun-info call)))
                        (not (or (fun-info-ir2-convert info)
                                 (ir1-attributep (fun-info-attributes info)
@@ -979,11 +1061,14 @@
          (:known
           (ltn-analyze-known-call node))))
       (cif (ltn-analyze-if node))
+      (jump-table (ltn-analyze-jump-table node))
       (creturn) ;; delay to FLUSH-FULL-CALL-TAIL-TRANSFERS
       ((or bind entry))
       (exit (ltn-analyze-exit node))
       (cset (ltn-analyze-set node))
       (cast (ltn-analyze-cast node))
+      (enclose)
+      (cdynamic-extent (ltn-analyze-dynamic-extent node))
       (mv-combination
        (ecase (basic-combination-kind node)
          (:local
