@@ -217,6 +217,9 @@
                  (setf (block-reoptimize (node-block node)) t)
                  (reoptimize-component (node-component node) :maybe))
                t)
+             (change-return-type (node type)
+               (setf (node-derived-type node) type)
+               (setf (lvar-%derived-type (node-lvar node)) nil))
              (cut-node (node)
                "Try to cut a node to width. The primary return value is
                 whether we managed to cut (cleverly), and the second whether
@@ -241,10 +244,7 @@
                              (t
                               (change-ref-leaf node (find-constant new-value)
                                                :recklessly t)
-                              (let ((lvar (node-lvar node)))
-                                (setf (lvar-%derived-type lvar)
-                                      (and (lvar-has-single-use-p lvar)
-                                           (make-values-type (list (ctype-of new-value))))))
+                              (change-return-type node (make-values-type (list (ctype-of new-value))))
                               (setf (block-reoptimize (node-block node)) t)
                               (reoptimize-component (node-component node) :maybe)
                               (values t t)))))))
@@ -301,13 +301,36 @@
                                  ;; Can't rely on REOPTIMIZE-NODE, as it may neve get reoptimized.
                                  ;; But the outer functions don't want the type to get
                                  ;; widened and their VOPs may never be applied.
-                                 (setf (node-derived-type node)
-                                       (fun-type-returns (global-ftype (if (eq name t)
-                                                                           fun-name
-                                                                           name))))
-                                 (setf (lvar-%derived-type (node-lvar node)) nil)
+                                 (change-return-type node
+                                                     (fun-type-returns (global-ftype (if (eq name t)
+                                                                                         fun-name
+                                                                                         name))))
                                  (ir1-optimize-combination node))
-                               (values t did-something over-wide)))))))))
+                               (values t did-something over-wide)))))))
+                 (cast
+                  ;; Cut (logand (+ x 1) m), which is (logand (the integer (+ x 1)) m),
+                  ;; and X can only be an integer for that to be true.
+                  (when (eq (cast-type-to-check node)
+                            (specifier-type 'integer))
+                    (let (did-something)
+                      (do-uses (combination (cast-value node))
+                        (when (and (or (combination-matches* '(+ -) '(* *) combination)
+                                       (combination-matches* '(%negate) '(*) combination))
+                                   (almost-immediately-used-p (node-lvar combination) combination
+                                                              :flushable t))
+                          (destructuring-bind (a &optional b) (combination-args combination)
+                            (when (or (not (types-equal-or-intersect (lvar-type a)
+                                                                     #1=(specifier-type '(or ratio (complex rational)))))
+                                      (not (and b
+                                                (types-equal-or-intersect (lvar-type b) #1#))))
+                              (when (cut-node combination)
+                                (setf did-something t))))))
+                      (when did-something
+                        (change-return-type node
+                                            (if (type-single-value-p (node-derived-type node))
+                                                (values-specifier-type '(values integer &optional))
+                                                (values-specifier-type '(values integer)))))
+                      nil)))))
              (cut-lvar (lvar &key head
                         &aux did-something must-insert over-wide)
                "Cut all the LVAR's use nodes. If any of them wasn't handled
@@ -369,8 +392,9 @@
           (return-from best-modular-version
             (values (car ugt) :untagged (cdr ugt))))))))
 
-(defoptimizer (logand optimizer) ((x y) node)
-  (let ((result-type (single-value-type (node-derived-type node))))
+(defoptimizer (logand optimizer) ((x y) node &optional result-type)
+  (let ((result-type (or result-type
+                         (single-value-type (node-derived-type node)))))
     (multiple-value-bind (low high)
         (integer-type-numeric-bounds result-type)
       (when (and (numberp low)
@@ -401,6 +425,11 @@
               )))))))
 
 (setf (fun-info-optimizer (fun-info-or-lose 'logandc2)) #'logand-optimizer-optimizer)
+(setf (fun-info-optimizer (fun-info-or-lose 'logtest))
+      (lambda (node)
+        (let ((type (logand-derive-type-optimizer node)))
+          (when type
+           (logand-optimizer-optimizer node type)))))
 
 (defoptimizer (mask-signed-field optimizer) ((width x) node)
   (let ((result-type (single-value-type (node-derived-type node))))
@@ -747,7 +776,7 @@
         (destructuring-bind (q r) (values-type-required res)
           (make-values-type  (list (%two-arg-derive-type q
                                                          (specifier-type `(eql ,(ldb (byte sb-vm:n-word-bits 0) -1)))
-                                                         #'logand-derive-type-aux #'logand)
+                                                         #'logand-derive-type-aux)
                                    r))))))
 
   (deftransform sb-vm::truncate-mod64 ((n d) * * :node node)
