@@ -213,8 +213,9 @@
       ;; using the process signal mask.
       #+partial-sw-int-avoidance (inst call (ea (make-fixup 'synchronous-trap :assembly-routine)))
       #-partial-sw-int-avoidance (progn
-      #+int1-breakpoints (inst icebp)
-      #-int1-breakpoints (inst break pending-interrupt-trap))
+      ;; *NOT* ud2-breakpoints means *do* use UD2 for pseudo-atomic, as a special case.
+      #-ud2-breakpoints (inst ud2)
+      #+ud2-breakpoints (inst break pending-interrupt-trap))
       OUT)))
 
 (defmacro define-allocator (name &body body &aux (g (cdr (assoc :generator body))))
