@@ -329,7 +329,9 @@
   (:policy :fast-safe)
   (:translate pointerp)
   (:generator 3
-    (if (location= temp value) (inst sub :dword value 3) (inst lea :dword temp (ea -3 value)))
+    ;; Since TEST will examine only the low 2 bits, it doesn't matter if we flip just
+    ;; those bits, or the low dword. The latter can be done without an immediate operand.
+    (if (location= temp value) (inst not :dword value) (inst lea :dword temp (ea -3 value)))
     (inst test :byte temp #b11)))
 
 ;; A fixnum or single-digit bignum satisfies signed-byte-64-p
@@ -615,7 +617,17 @@
                        (inst jmp :nz nope)))
                 ;; Check if the remaining high bits are zero
                 (let ((bits (dpb 0 (byte left 0) -1)))
-                  (cond ((< left 32)
+                  (cond (unsigned-p
+                         (inst jmp :l yep)
+                         (cond ((< left 32)
+                                (inst test :qword (ea (- other-pointer-lowtag) value temp n-word-bytes) bits))
+                               (t
+                                (inst test :dword (ea (+ (- other-pointer-lowtag)
+                                                         (/ n-word-bytes 2))
+                                                      value temp n-word-bytes)
+                                      (ash bits -32)))))
+
+                        ((< left 32)
                          (inst mov last-digit (ea (- other-pointer-lowtag) value temp n-word-bytes))
                          (inst jmp :l fixnum)
                          (inst test last-digit bits))
@@ -661,7 +673,11 @@
                          (change-vop-flags vop '(:nz))))
                       (t
                        (if (location= temp value)
-                           (inst sub :dword value ,lowtag)
+                           ;; Similarly with POINTERP, this might avoid an immediate operand.
+                           ;; (Seems like we rarely if ever get here with LOCATION= though)
+                           ,(if (= (symbol-value lowtag) #b1111)
+                                '(inst not :dword value)
+                                `(inst xor :dword value ,lowtag))
                            (inst lea :dword temp (ea (- ,lowtag) value)))
                        (inst test :byte temp lowtag-mask))))))))
   (define functionp fun-pointer-lowtag)
